/*-
 * Copyright (c) 2009 Oleksandr Tymoshenko <gonzo@FreeBSD.org>
 * Copyright (c) 2013 The FreeBSD Foundation
 * All rights reserved.
 *
 * Portions of this software were developed by Oleksandr Rybalko
 * under sponsorship from the FreeBSD Foundation.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice unmodified, this list of conditions, and the following
 *    disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/param.h>
#include <sys/systm.h>

#include <sys/bus.h>
#include <sys/interrupt.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/rman.h>

#include <vm/vm.h>
#include <vm/pmap.h>
#include <vm/vm_extern.h>

#include <dev/fdt/fdt_common.h>
#include <dev/ofw/openfirm.h>
#include <dev/ofw/ofw_bus.h>
#include <dev/ofw/ofw_bus_subr.h>

#include <machine/bus.h>
#include <machine/cpu.h>
#include <machine/pmap.h>
#include <machine/fdt.h>
#include <sys/endian.h>

#include <dev/spibus/spi.h>
#include <dev/spibus/spibusvar.h>
#include "spibus_if.h"

#include <arm/freescale/imx/imx51_ccmvar.h>
#include <arm/freescale/imx/imx_spireg.h>
#include <sys/gpio.h>
#include <dev/gpio/gpiobusvar.h>
#include "gpio_if.h"

#undef IMX_SPI_DEBUG
#ifdef IMX_SPI_DEBUG
#define dprintf printf
#else
#define dprintf(x, arg...)
#endif

/*
 * register space access macros
 */
#define SPI_WRITE(sc, reg, val)	do {					\
		bus_write_4(sc->sc_mem_res, (reg), (val)); 		\
	} while (0)
#define SPI_READ(sc, reg)	 bus_read_4(sc->sc_mem_res, (reg))
#define SPI_SET_BITS(sc, reg, bits)					\
	    SPI_WRITE(sc, reg, SPI_READ(sc, (reg)) | (bits))
#define SPI_CLEAR_BITS(sc, reg, bits)					\
	    SPI_WRITE(sc, reg, SPI_READ(sc, (reg)) & ~(bits))

struct imx_spi_softc {
	device_t	sc_dev;
	struct resource	*sc_mem_res;
	int		sc_clk;
	int		post_div;
	int		pre_div;
	int8_t		cs_gpio[4];
	int8_t		cs_gpio_inv[4];	/* Invert CS (Active is Low) */
};

static int	imx_spi_probe(device_t);
static int	imx_spi_attach(device_t);
static int	imx_spi_detach(device_t);
static int	imx_spi_wait(struct imx_spi_softc *);
static void	imx_spi_chip_activate(struct imx_spi_softc *, int);
static void	imx_spi_chip_deactivate(struct imx_spi_softc *, int);
static int	imx_spi_transfer(device_t, device_t, struct spi_command *);
static int	imx_spi_txrx(struct imx_spi_softc *sc,
    struct spibus_ivar *devi, int l, uint8_t *tx, uint8_t *rx);

static void
gpio_api_set(int p, int v)
{
	device_t dev;
	int c;

	if (p == -1)
		return;

	v = (v) ? 1 : 0;
	c = p / 32; /* 32 GPIO per controller */
	p %= 32; /* 120 -> 24, gpioc3-pin24 */
	dev = devclass_get_device(devclass_find("gpio"), c);
	if (dev == NULL)
		return;

	GPIO_PIN_SETFLAGS(dev, p, GPIO_PIN_OUTPUT);
	GPIO_PIN_SET(dev, p, v);
}


static int
imx_spi_probe(device_t dev)
{

	if (!ofw_bus_is_compatible(dev, "fsl,imx51-ecspi"))
		return (ENXIO);

	device_set_desc(dev, "Freescale i.MXxxx sCSPI controller");
	return (BUS_PROBE_DEFAULT);
}

static int
imx_spi_attach(device_t dev)
{
	struct imx_spi_softc *sc;
	int rid;

	sc = device_get_softc(dev);
	sc->sc_dev = dev;
        rid = 0;
	sc->sc_mem_res = bus_alloc_resource_any(dev, SYS_RES_MEMORY, &rid,
	    RF_ACTIVE);
	if (!sc->sc_mem_res) {
		device_printf(dev, "Could not map memory\n");
		return (ENXIO);
	}

	int i;
	for (i = 0x08; i <= 0x20; i += 4)
		printf("\t%02x:	%08x\n", i, SPI_READ(sc, i));


	if (imx_spi_wait(sc)) {
		bus_release_resource(dev, SYS_RES_MEMORY, 0, sc->sc_mem_res);
		return (EBUSY);
	}

	sc->sc_clk = imx51_get_clock(IMX51CLK_IPG_CLK_ROOT);
	/* XXX: hardcoded, CS0 - active high, CS1 - active low */
	sc->cs_gpio[0] = 120;
	sc->cs_gpio[1] = 121;
	sc->cs_gpio[2] = -1;
	sc->cs_gpio[3] = -1;
	sc->cs_gpio_inv[0] = 1;	/* Invert CS (Active is High) */

	device_add_child(dev, "spibus", -1);
	return (bus_generic_attach(dev));
}

static int
imx_spi_detach(device_t dev)
{
	struct imx_spi_softc *sc;

	sc = device_get_softc(dev);
	if (sc->sc_mem_res)
		bus_release_resource(dev, SYS_RES_MEMORY, 0, sc->sc_mem_res);

	return (0);
}

static void
imx_spi_chip_activate(struct imx_spi_softc *sc, int cs)
{
	int i;

	for (i = 0; i < 4; i ++)
		if (i == cs)
			gpio_api_set(sc->cs_gpio[i],
			    (sc->cs_gpio_inv[i]) ? 1 : 0);
		else
			gpio_api_set(sc->cs_gpio[i],
			    (sc->cs_gpio_inv[i]) ? 0 : 1);

}

static void
imx_spi_chip_deactivate(struct imx_spi_softc *sc, int cs)
{
	int i;

	/* Put all CSx to high */
	for (i = 0; i < 4; i ++)
		gpio_api_set(sc->cs_gpio[i],
		    (sc->cs_gpio_inv[i]) ? 0 : 1);
}

static int
imx_spi_wait(struct imx_spi_softc *sc)
{
	int i = 100000;

	while (i-- && (SPI_READ(sc, IMX_SPI_STS) & IMX_SPI_STS_TC) == 0);

	if (i == 0) {
		printf("busy\n");
		return (1);
	}

	/* ACK TC bit */
	SPI_WRITE(sc, IMX_SPI_STS, IMX_SPI_STS_TC);

	return (0);
}
enum {
	IMX_SPI_READ=0,
	IMX_SPI_WRITE
};

static int
imx_spi_txrx(struct imx_spi_softc *sc, struct spibus_ivar *devi, int l,
    uint8_t *tx, uint8_t *rx)
{
	uint32_t reg, ctlreg;
	int i, saved;

	saved = l;
	ctlreg =
	    (((l * 8) - 1) << IMX_SPI_CTRL_BRSTLEN_SHIFT) |
	    ((devi->cs) << IMX_SPI_CTRL_CHAN_SHIFT) |
	    (sc->pre_div << IMX_SPI_CTRL_PRE_DIV_SHIFT) |
	    (sc->post_div << IMX_SPI_CTRL_POST_DIV_SHIFT) |
	    IMX_SPI_CTRL_MASTER_CH(devi->cs) |
	    IMX_SPI_CTRL_EN |
	    0;

	SPI_WRITE(sc, IMX_SPI_CTRL, ctlreg);
	SPI_WRITE(sc, IMX_SPI_CFG, 0);
	SPI_WRITE(sc, IMX_SPI_INT, 0);
	SPI_WRITE(sc, IMX_SPI_STS, IMX_SPI_STS_TC | IMX_SPI_STS_RO);

	if (tx) {
		if (l & 3) {
			hexdump(tx, MIN(5, l), NULL, 0);
			reg = 0;
			for (i = 0; i < (l & 3); i ++)
				reg = (reg << 8) | tx[i];
			tx += (l & 3);
			l &= 3;
//			printf("txU: 0x%08x\n", htole32(reg));
			SPI_WRITE(sc, IMX_SPI_TXDATA, htole32(reg));
		}

		if ((vm_offset_t)tx & 3) {
			/* If address is not 32bits alligned */
			for (i = 0; i < l; i += 4) {
				reg = (tx[i] << 24) |
				    (tx[i + 1] << 16) |
				    (tx[i + 2] << 8) |
				    (tx[i + 3]);
//				if (i == 0)
//					printf("tx0: 0x%08x, tx+[%02x,%02x,%02x,%02x]\n", htole32(reg), (tx[i]), (tx[i + 1]), (tx[i + 2]), (tx[i + 3]));
				SPI_WRITE(sc, IMX_SPI_TXDATA, htole32(reg));
			}
		} else {
			/* Aligned address/size, most common case for data */
			for (i = 0; i < (l >> 2); i ++) {
				reg = ((uint32_t *)tx)[i];
				SPI_WRITE(sc, IMX_SPI_TXDATA, htobe32(reg));
			}
		}
	} else {
			for (i = 0; i < l; i += 4) {
				reg = 0;
				SPI_WRITE(sc, IMX_SPI_TXDATA, reg);
			}
	}

	SPI_SET_BITS(sc, IMX_SPI_CTRL, IMX_SPI_CTRL_EN | IMX_SPI_CTRL_XCH);

	if (imx_spi_wait(sc))
		return (EBUSY);

	if (rx) {
		l = saved;
		if (l & 3) {
			reg = be32toh(SPI_READ(sc, IMX_SPI_RXDATA));
			printf("rxU: 0x%08x\n", reg);
			reg >>= ((4 - (l & 3)) * 8);
			for (i = 0; i < (l & 3) ; i ++) {
				rx[i] =  reg & 0xff;
				reg >>= 8;
			}
			rx += (l & 3);
			l -= (l & 3);
		}

		if ((vm_offset_t)rx & 3) {
			/* If address is not 32bits alligned */
			for (i = 0; i < l; i += 4) {
				reg = le32toh(SPI_READ(sc, IMX_SPI_RXDATA));
				rx[i + 3] = (reg >> 24) & 0xff;
				rx[i + 2] = (reg >> 16) & 0xff;
				rx[i + 1] = (reg >> 8)  & 0xff;
				rx[i    ] =  reg        & 0xff;
			}
		} else {
			/* Aligned address/size, most common case for data */
			for (i = 0; i < (l >> 2); i ++) {
				reg = be32toh(SPI_READ(sc, IMX_SPI_RXDATA));
				((uint32_t *)rx)[i] = reg;
			}
		}
	} else {
		/* Drain FIFO */
		for (i = 0; i < saved; i += 4)
			SPI_READ(sc, IMX_SPI_RXDATA);
	}

	return (0);
}

static int
imx_spi_calc_div(struct imx_spi_softc *sc, int clk)
{
	int i, post_div, pre_div;

	pre_div = post_div = 0;
	if (sc->sc_clk > clk) {
    		pre_div = sc->sc_clk / clk;
    		if (pre_div > 16) {
        		post_div = pre_div / 16;
        		pre_div = 15;
    		}
    		if (post_div != 0) {
        		for (i = 0; i < 16; i++) {
            			if ((1 << i) >= post_div)
                			break;
        		}
        		if (i == 16) {
            			device_printf(sc->sc_dev, "ERROR: Can't find "
            			    "correct dividers for the freq: %d\n", clk);
            			return (EIO);
        		}
        		post_div = i;
    		}
	} else if (sc->sc_clk == clk){
		/* attach sys_clk direct, w/o dividers */
		pre_div = 0;
		post_div = 0;
	} else {
    		device_printf(sc->sc_dev, "ERROR: SPI bus controller "
    		    "frequency to low %d Hz for %d Hz\n", sc->sc_clk, clk);
    		return (EIO);
	}

	sc->pre_div = pre_div;
	sc->post_div = post_div;
	return (0);
}

static int
imx_spi_transfer(device_t dev, device_t child, struct spi_command *cmd)
{
	struct imx_spi_softc *sc;
	struct spibus_ivar *devi;
	uint32_t reg;
	int error, len, tx;
	uint8_t *ptr;

	error = 0;
	sc = device_get_softc(dev);
	devi = SPIBUS_IVAR(child);

	if (devi->cs & ~3)
		/* Only CS1-CS4 CS */
		return (ENXIO);
	if ((cmd->tx_cmd_sz > 0) && (cmd->rx_cmd_sz > 0))
		return (EIO);
	if ((cmd->tx_data_sz > 0) && (cmd->rx_data_sz > 0))
		return (EIO);

	imx_spi_chip_activate(sc, devi->cs);

	if (imx_spi_calc_div(sc, 25000000) != 0)
		return (EIO);

	/* Reset controller */
	reg = SPI_READ(sc, IMX_SPI_CTRL);
	SPI_WRITE(sc, IMX_SPI_CTRL, 0);
	SPI_WRITE(sc, IMX_SPI_CTRL, reg | IMX_SPI_CTRL_EN);

	/*
	 * Transfer command
	 */
	imx_spi_txrx(sc, devi, cmd->tx_cmd_sz, cmd->tx_cmd, NULL);

	if (cmd->tx_cmd_sz < 16 && devi->cs == 1) {
		hexdump(cmd->tx_cmd, cmd->tx_cmd_sz, "tx cmd:", 0);
	}
	/*
	 * Transfer/Receive data
	 */
	if (cmd->rx_data_sz > 0) {
		tx = 0;
		len = cmd->rx_data_sz;
		ptr = cmd->rx_data;
	} else {
		tx = 1;
		len = cmd->tx_data_sz;
		ptr = cmd->tx_data;
	}

	while (len > 0) {
		imx_spi_txrx(sc, devi, MIN(256,len),
		    (tx) ? ptr : NULL,
		    (tx) ? NULL : ptr);
		len -= 256;
		ptr += 256;
	}

//	if (cmd->rx_data_sz > 0 && devi->cs == 1) {
//		hexdump(cmd->rx_data, cmd->rx_data_sz, NULL, 0);
//	}

	imx_spi_chip_deactivate(sc, devi->cs);

	return (error);
}

static phandle_t imx_spi_get_node(device_t, device_t);

static phandle_t
imx_spi_get_node(device_t bus, device_t dev)
{
	/*
	 * Share controller node with spibus device
	 */
	return ofw_bus_get_node(bus);
}

static device_method_t imx_spi_methods[] = {
	/* Device interface */
	DEVMETHOD(device_probe,		imx_spi_probe),
	DEVMETHOD(device_attach,	imx_spi_attach),
	DEVMETHOD(device_detach,	imx_spi_detach),
	/* SPI methods */
	DEVMETHOD(spibus_transfer,	imx_spi_transfer),
	/* OFW methods */
	DEVMETHOD(ofw_bus_get_node,	imx_spi_get_node),
	DEVMETHOD_END
};

static driver_t imx_spi_driver = {
	"spi",
	imx_spi_methods,
	sizeof(struct imx_spi_softc),
};

static devclass_t imx_spi_devclass;

DRIVER_MODULE(spi, simplebus, imx_spi_driver, imx_spi_devclass, 0, 0);
DRIVER_MODULE(spibus, spi, spibus_driver, spibus_devclass, 0, 0);
